"""Ridge model for classification"""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
from scipy.optimize import minimize
import sklearn.metrics as skm2
from .ridge import Ridge
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import ClassifierMixin
from scipy.special import logsumexp


class RidgeClassifier(Ridge, ClassifierMixin):
    """Ridge Classification model class with 2 regularization parameters derived from class Ridge
    
       Parameters
       ----------
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not 
           (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
               no clustering)
       cluster_encode: bool
           defines how the variable containing clusters is treated (default is one-hot)
           if `False`, then labels are used, without one-hot encoding
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training   
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       lambda1: float
           regularization parameter on direct link
       lambda2: float
           regularization parameter on hidden layer
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        lambda1=0.1,
        lambda2=0.1,
        seed=123,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            lambda1=lambda1,
            lambda2=lambda2,
            seed=seed,
        )

        self.type_fit = "classification"

    def loglik(self, X, Y, **kwargs):
        """Log-likelihood for training data (X, Y).
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        Y: array-like, shape = [n_samples]
               One-hot encode target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set or self.obj.fit
               
        Returns
        -------
        """

        def loglik_grad_hess(
            Y, X, B, XB, hessian=True, **kwargs
        ):
            # nobs, n_classes
            n, K = Y.shape

            # total number of covariates
            p = X.shape[1]

            # initial number of covariates
            init_p = p - self.n_hidden_features

            max_double = 709.0
            XB[XB > max_double] = max_double
            exp_XB = np.exp(XB)
            probs = exp_XB / exp_XB.sum(axis=1)[:, None]

            # gradient -----
            # (Y - p) -> (n, K)
            # X -> (n, p)
            # (K, n) %*% (n, p) -> (K, p)
            if hessian is False:
                grad = -np.dot((Y - probs).T, X) / n
                grad += (
                    self.lambda1
                    * B[0:init_p, :].sum(axis=0)[:, None]
                )
                grad += (
                    self.lambda2
                    * B[init_p:p, :].sum(axis=0)[:, None]
                )

                return grad.flatten()

            # hessian -----
            if hessian is True:
                Kp = K * p
                hess = np.zeros((Kp, Kp), float)
                for k1 in range(K):
                    x_index = range(k1 * p, (k1 + 1) * p)
                    for k2 in range(k1, K):
                        y_index = range(
                            k2 * p, (k2 + 1) * p
                        )
                        H_sub = (
                            -np.dot(
                                X.T,
                                (
                                    probs[:, k1]
                                    * probs[:, k2]
                                )[:, None]
                                * X,
                            )
                            / n
                        )  # do not store
                        hess[
                            np.ix_(x_index, y_index)
                        ] = hess[
                            np.ix_(y_index, x_index)
                        ] = H_sub

                return hess + (
                    self.lambda1 + self.lambda2
                ) * np.identity(Kp)

        # total number of covariates
        p = X.shape[1]

        # initial number of covariates
        init_p = p - self.n_hidden_features

        # log-likelihood (1st return)
        def loglik_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            # (n, K)
            XB = mo.safe_sparse_dot(X, B)

            res = -(
                np.sum(Y * XB, axis=1) - logsumexp(XB)
            ).mean()

            res += (
                0.5
                * self.lambda1
                * mo.squared_norm(B[0:init_p, :])
            )
            res += (
                0.5
                * self.lambda2
                * mo.squared_norm(B[init_p:p, :])
            )

            return res

        # gradient of log-likelihood
        def grad_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            return loglik_grad_hess(
                Y=Y,
                X=X,
                B=B,
                XB=mo.safe_sparse_dot(X, B),
                hessian=False,
                **kwargs
            )

        # hessian of log-likelihood
        def hessian_func(x):
            # (p, K)
            B = x.reshape(Y.shape[1], p).T

            return loglik_grad_hess(
                Y=Y,
                X=X,
                B=B,
                XB=mo.safe_sparse_dot(X, B),
                hessian=True,
                **kwargs
            )

        return loglik_func, grad_func, hessian_func

    # newton-cg
    # L-BFGS-B
    def fit(self, X, y, solver="L-BFGS-B", **kwargs):
        """Fit Ridge model to training data (X, y).
           for beta: regression coeffs (beta11, ..., beta1p, ..., betaK1, ..., betaKp)
           for K classes and p covariates.

        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        y: array-like, shape = [n_samples]
               Target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set or self.obj.fit
               
        Returns
        -------
        self: object
        """

        assert mx.is_factor(
            y
        ), "y must contain only integers"

        output_y, scaled_Z = self.cook_training_set(
            y=y, X=X, **kwargs
        )

        self.n_classes = len(np.unique(y))

        Y = mo.one_hot_encode2(output_y, self.n_classes)

        # optimize for beta, minimize self.loglik (maximize loglik) -----
        loglik_func, grad_func, hessian_func = self.loglik(
            X=scaled_Z, Y=Y
        )

        if solver == "L-BFGS-B":
            self.beta = minimize(
                fun=loglik_func,
                x0=np.zeros(
                    scaled_Z.shape[1] * self.n_classes
                ),
                jac=grad_func,
                method=solver,
            ).x

        if solver in ("Newton-CG", "trust-ncg"):
            self.beta = minimize(
                fun=loglik_func,
                x0=np.zeros(
                    scaled_Z.shape[1] * self.n_classes
                ),
                jac=grad_func,
                hess=hessian_func,
                method=solver,
            ).x

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions: {array-like}
        """

        return np.argmax(
            self.predict_proba(X, **kwargs), axis=1
        )

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        probability estimates for test data: {array-like}        
        """
        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

        else:

            Z = self.cook_test_set(X, **kwargs)

        ZB = mo.safe_sparse_dot(
            Z,
            self.beta.reshape(
                self.n_classes,
                X.shape[1]
                + self.n_hidden_features
                + self.n_clusters,
            ).T,
        )
        exp_ZB = np.exp(ZB)

        return exp_ZB / exp_ZB.sum(axis=1)[:, None]

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set covariates X and response y. """

        preds = self.predict(X)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



from .ridgeClassifier import RidgeClassifier
from .ridgeRegressor import RidgeRegressor
from .ridgeClassifierMtask import RidgeClassifierMtask

__all__ = ["RidgeClassifier", "RidgeRegressor", "RidgeClassifierMtask"]



from .base.base import Base
from .base.baseRegressor import BaseRegressor
from .boosting.adaBoostClassifier import AdaBoostClassifier
from .custom.customClassifier import CustomClassifier
from .custom.customRegressor import CustomRegressor
from .mts import MTS
from .randombag.randomBagClassifier import (
    RandomBagClassifier,
)
from .ridge.ridgeClassifier import RidgeClassifier
from .ridge.ridgeRegressor import RidgeRegressor
from .ridge.ridgeClassifierMtask import RidgeClassifierMtask

# from .rnn.rnnRegressor import RNNRegressor
# from .rnn.rnnClassifier import RNNClassifier
from .rvfl.bayesianrvflRegressor import (
    BayesianRVFLRegressor,
)
from .rvfl.bayesianrvfl2Regressor import (
    BayesianRVFL2Regressor,
)


__all__ = [
    "AdaBoostClassifier",
    "Base",
    "BaseRegressor",
    "BayesianRVFLRegressor",
    "BayesianRVFL2Regressor",
    "CustomClassifier",
    "CustomRegressor",
    "RandomBagClassifier",
    "RidgeRegressor",
    "RidgeClassifier",
    "RidgeClassifierMtask",
    #    "RNNRegressor",
    #    "RNNClassifier",
    "MTS",
]



"""Base class."""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
from functools import partial
from sklearn.base import BaseEstimator
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
)
from ..utils import activations as ac
from ..utils import memoize
from ..utils import matrixops as mo
from ..utils import misc as mx
from ..simulation import nodesimulation as ns
from ..simulation import rowsubsampling as rs


class Base(BaseEstimator):
    """Base model with direct link and nonlinear activation.
        
       Parameters
       ----------
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or 
           not (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for type_clust='kmeans' or type_clust='gmm' 
           clustering (could be 0: no clustering)
       cluster_encode: bool
           defines how the variable containing clusters is treated (default is one-hot)
           if `False`, then labels are used, without one-hot encoding
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training    
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform', clustering and dropout
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=True,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        seed=123,
    ):

        # input checks -----

        assert activation_name in (
            "relu",
            "tanh",
            "sigmoid",
            "prelu",
            "elu",
        ), "'activation_name' must be in ('relu', 'tanh', 'sigmoid','prelu', 'elu')"

        assert nodes_sim in (
            "sobol",
            "hammersley",
            "uniform",
            "halton",
        ), "'nodes_sim' must be in ('sobol', 'hammersley', 'uniform', 'halton')"

        assert type_clust in (
            "kmeans",
            "gmm",
        ), "'type_clust' must be in ('kmeans', 'gmm')"

        assert (len(type_scaling) == 3) & all(
            type_scaling[i] in ("minmax", "std")
            for i in range(len(type_scaling))
        ), "'type_scaling' must have length 3, and available scaling methods are 'minmax' scaling and standardization ('std')"

        assert (col_sample >= 0) & (
            col_sample <= 1
        ), "'col_sample' must be comprised between 0 and 1 (both included)"

        # activation function -----

        activation_options = {
            "relu": ac.relu,
            "tanh": np.tanh,
            "sigmoid": ac.sigmoid,
            "prelu": partial(ac.prelu, a=a),
            "elu": partial(ac.elu, a=a),
        }

        self.n_hidden_features = n_hidden_features
        self.activation_name = activation_name
        self.a = a
        self.activation_func = activation_options[
            activation_name
        ]
        self.nodes_sim = nodes_sim
        self.bias = bias
        self.seed = seed
        self.dropout = dropout
        self.direct_link = direct_link
        self.cluster_encode = cluster_encode
        self.type_clust = type_clust
        self.type_scaling = type_scaling
        self.col_sample = col_sample
        self.row_sample = row_sample
        self.index_col = None
        self.index_row = True
        self.n_clusters = n_clusters
        self.clustering_obj = None
        self.clustering_scaler = None
        self.nn_scaler = None
        self.scaler = None
        self.encoder = None
        self.W = None
        self.X = None
        self.y = None
        self.y_mean = None
        self.beta = None

    # "preprocessing" methods to be inherited -----

    def encode_clusters(
        self, X=None, predict=False, **kwargs
    ):  #
        """ Create new covariates with kmeans or GMM clustering. 

        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        predict: boolean
            is False on training set and True on test set
        
        **kwargs: 
            additional parameters to be passed to the 
            clustering method  
            
        Returns
        -------
        clusters' matrix, one-hot encoded: {array-like}        
        """

        if X is None:
            X = self.X

        if predict == False:  # encode training set

            # scale input data before clustering
            self.clustering_scaler, scaled_X = mo.scale_covariates(
                X, choice=self.type_scaling[2]
            )

            if self.type_clust == "kmeans":

                self.clustering_obj, X_kmeans = mo.cluster_covariates(
                    scaled_X,
                    self.n_clusters,
                    self.seed,
                    type_clust="kmeans",
                    **kwargs
                )

                if self.cluster_encode == True:

                    return mo.one_hot_encode(
                        X_kmeans, self.n_clusters
                    )

                return X_kmeans

            if self.type_clust == "gmm":

                self.clustering_obj, X_gmm = mo.cluster_covariates(
                    scaled_X,
                    self.n_clusters,
                    self.seed,
                    type_clust="gmm",
                    **kwargs
                )

                if self.cluster_encode == True:

                    return mo.one_hot_encode(
                        X_gmm, self.n_clusters
                    )

                return X_gmm

        # if predict == True, encode test set
        X_clustered = self.clustering_obj.predict(
            self.clustering_scaler.transform(X)
        )

        if self.cluster_encode == True:

            return mo.one_hot_encode(
                X_clustered, self.n_clusters
            )

        return X_clustered

    def create_layer(self, scaled_X, W=None):
        """ Create hidden layer. """

        n_features = scaled_X.shape[1]

        if (
            self.bias is False
        ):  # no bias term in the hidden layer

            if W is None:

                if self.nodes_sim == "sobol":
                    try:  # try cpp version
                        self.W = ns.generate_sobol_cpp(
                            n_dims=n_features,
                            n_points=self.n_hidden_features,
                        )
                    except:  # python version
                        self.W = ns.generate_sobol2(
                            n_dims=n_features,
                            n_points=self.n_hidden_features,
                        )

                if self.nodes_sim == "hammersley":
                    self.W = ns.generate_hammersley(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                    )

                if self.nodes_sim == "uniform":
                    self.W = ns.generate_uniform(
                        n_dims=n_features,
                        n_points=self.n_hidden_features,
                        seed=self.seed,
                    )

                if self.nodes_sim == "halton":
                    try:  # try cpp version
                        self.W = ns.generate_halton_cpp(
                            n_dims=n_features,
                            n_points=self.n_hidden_features,
                        )
                    except:  # python version
                        self.W = ns.generate_halton(
                            n_dims=n_features,
                            n_points=self.n_hidden_features,
                        )

                assert (
                    scaled_X.shape[1] == self.W.shape[0]
                ), "check dimensions of covariates X and matrix W"

                return mo.dropout(
                    x=self.activation_func(
                        np.dot(scaled_X, self.W)
                    ),
                    drop_prob=self.dropout,
                    seed=self.seed,
                )

            # W is not none
            assert (
                scaled_X.shape[1] == W.shape[0]
            ), "check dimensions of covariates X and matrix W"

            # self.W = W
            return mo.dropout(
                x=self.activation_func(np.dot(scaled_X, W)),
                drop_prob=self.dropout,
                seed=self.seed,
            )

        # with bias term in the hidden layer
        if W is None:

            n_features_1 = n_features + 1

            if self.nodes_sim == "sobol":
                try:  # try cpp version
                    self.W = ns.generate_sobol_cpp(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )
                except:  # python version
                    self.W = ns.generate_sobol2(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )

            if self.nodes_sim == "hammersley":
                self.W = ns.generate_hammersley(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                )

            if self.nodes_sim == "uniform":
                self.W = ns.generate_uniform(
                    n_dims=n_features_1,
                    n_points=self.n_hidden_features,
                    seed=self.seed,
                )

            if self.nodes_sim == "halton":
                try:  # try cpp version
                    self.W = ns.generate_halton_cpp(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )
                except:  # python version
                    self.W = ns.generate_halton(
                        n_dims=n_features_1,
                        n_points=self.n_hidden_features,
                    )

            return mo.dropout(
                x=self.activation_func(
                    np.dot(
                        mo.cbind(
                            np.ones(scaled_X.shape[0]),
                            scaled_X,
                        ),
                        self.W,
                    )
                ),
                drop_prob=self.dropout,
                seed=self.seed,
            )

        # W is not None
        # self.W = W
        return mo.dropout(
            x=self.activation_func(
                np.dot(
                    mo.cbind(
                        np.ones(scaled_X.shape[0]), scaled_X
                    ),
                    W,
                )
            ),
            drop_prob=self.dropout,
            seed=self.seed,
        )

    def cook_training_set(
        self, y=None, X=None, W=None, **kwargs
    ):
        """ Create new data for training set, with hidden layer, center the response. """

        # either X and y are stored or not
        # assert ((y is None) & (X is None)) | ((y is not None) & (X is not None))
        if self.n_hidden_features > 0:  # has a hidden layer

            assert len(self.type_scaling) >= 2, ""

        # center y
        if mx.is_factor(y) == False:  # regression

            if y is None:
                self.y_mean, centered_y = mo.center_response(
                    self.y
                )
            else:  # keep
                self.y_mean, centered_y = mo.center_response(
                    y
                )

        if X is None:

            if self.col_sample == 1:

                input_X = self.X

            else:

                n_features = self.X.shape[1]
                new_n_features = int(
                    np.ceil(n_features * self.col_sample)
                )
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features),
                    size=new_n_features,
                    replace=False,
                )
                self.index_col = index_col
                input_X = self.X[:, self.index_col]

        else:  # X is not None

            if self.col_sample == 1:

                input_X = X

            else:

                n_features = X.shape[1]
                new_n_features = int(
                    np.ceil(n_features * self.col_sample)
                )
                assert (
                    new_n_features >= 1
                ), "check class attribute 'col_sample' and the number of covariates provided for X"
                np.random.seed(self.seed)
                index_col = np.random.choice(
                    range(n_features),
                    size=new_n_features,
                    replace=False,
                )
                self.index_col = index_col
                input_X = X[:, self.index_col]

        if (
            self.n_clusters <= 0
        ):  # data without any clustering: self.n_clusters is None -----

            if (
                self.n_hidden_features > 0
            ):  # with hidden layer

                self.nn_scaler, scaled_X = mo.scale_covariates(
                    input_X, choice=self.type_scaling[1]
                )

                if W is None:
                    Phi_X = self.create_layer(scaled_X)
                else:
                    Phi_X = self.create_layer(scaled_X, W=W)

                if self.direct_link == True:
                    Z = mo.cbind(input_X, Phi_X)
                else:
                    Z = Phi_X

                self.scaler, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )

            else:  # no hidden layer

                Z = input_X

                self.scaler, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )

        else:  # data with clustering: self.n_clusters is not None -----

            augmented_X = mo.cbind(
                input_X,
                self.encode_clusters(input_X, **kwargs),
            )

            if (
                self.n_hidden_features > 0
            ):  # with hidden layer

                self.nn_scaler, scaled_X = mo.scale_covariates(
                    augmented_X, choice=self.type_scaling[1]
                )

                if W is None:
                    Phi_X = self.create_layer(scaled_X)
                else:
                    Phi_X = self.create_layer(scaled_X, W=W)

                if self.direct_link == True:
                    Z = mo.cbind(augmented_X, Phi_X)
                else:
                    Z = Phi_X

                self.scaler, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )

            else:  # no hidden layer

                Z = augmented_X

                self.scaler, scaled_Z = mo.scale_covariates(
                    Z, choice=self.type_scaling[0]
                )

        # Returning model inputs -----

        # y is subsampled
        if self.row_sample < 1:

            n, p = Z.shape

            if y is None:

                self.index_row = rs.subsample(
                    y=self.y,
                    row_sample=self.row_sample,
                    seed=self.seed,
                )

            else:

                self.index_row = rs.subsample(
                    y=y,
                    row_sample=self.row_sample,
                    seed=self.seed,
                )

            n_row_sample = len(self.index_row)

            if mx.is_factor(y) == False:  # regression

                return (
                    centered_y[self.index_row].reshape(
                        n_row_sample
                    ),
                    self.scaler.transform(
                        Z[self.index_row, :].reshape(
                            n_row_sample, p
                        )
                    ),
                )

            # classification
            return (
                y[self.index_row].reshape(n_row_sample),
                self.scaler.transform(
                    Z[self.index_row, :].reshape(
                        n_row_sample, p
                    )
                ),
            )

        # y is not subsampled
        if mx.is_factor(y) == False:  # regression

            return (centered_y, self.scaler.transform(Z))

        # classification
        return (y, self.scaler.transform(Z))

    def cook_test_set(self, X, **kwargs):
        """ Transform data from test set, with hidden layer. """

        if (
            self.n_clusters == 0
        ):  # data without clustering: self.n_clusters is None -----

            if (
                self.n_hidden_features > 0
            ):  # if hidden layer

                if self.col_sample == 1:

                    scaled_X = self.nn_scaler.transform(X)

                else:

                    scaled_X = self.nn_scaler.transform(
                        X[:, self.index_col]
                    )

                Phi_X = self.create_layer(scaled_X, self.W)

                if self.direct_link == True:

                    return self.scaler.transform(
                        mo.cbind(scaled_X, Phi_X)
                    )

                # when self.direct_link == False
                return self.scaler.transform(Phi_X)

            # if no hidden layer # self.n_hidden_features == 0
            return self.scaler.transform(X)

        # data with clustering: self.n_clusters > 0 -----
        if self.col_sample == 1:

            predicted_clusters = self.encode_clusters(
                X=X, predict=True, **kwargs
            )
            augmented_X = mo.cbind(X, predicted_clusters)

        else:

            predicted_clusters = self.encode_clusters(
                X=X[:, self.index_col],
                predict=True,
                **kwargs
            )
            augmented_X = mo.cbind(
                X[:, self.index_col], predicted_clusters
            )

        if self.n_hidden_features > 0:  # if hidden layer

            scaled_X = self.nn_scaler.transform(augmented_X)
            Phi_X = self.create_layer(scaled_X, self.W)

            if self.direct_link == True:
                return self.scaler.transform(
                    mo.cbind(augmented_X, Phi_X)
                )

            return self.scaler.transform(Phi_X)

        # if no hidden layer
        return self.scaler.transform(augmented_X)



import numpy as np

from .memoize import memoize
from scipy import sparse
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
)
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture


# column bind
def cbind(x, y):

    # if len(x.shape) == 1 or len(y.shape) == 1:
    return np.column_stack((x, y))


# center... response
@memoize
def center_response(y):
    y_mean = np.mean(y)
    return y_mean, (y - y_mean)


# cluster the covariates
def cluster_covariates(
    X, n_clusters, seed, type_clust="kmeans", **kwargs
):

    if type_clust == "kmeans":

        kmeans = KMeans(
            n_clusters=n_clusters,
            random_state=seed,
            **kwargs
        )
        kmeans.fit(X)

        return kmeans, kmeans.predict(X)

    if type_clust == "gmm":

        gmm = GaussianMixture(
            n_components=n_clusters,
            random_state=seed,
            **kwargs
        )
        gmm.fit(X)

        return gmm, gmm.predict(X)


# computes t(x)%*%y
def crossprod(x, y=None):
    # assert on dimensions
    if y is None:
        return np.dot(x.transpose(), x)
    else:
        return np.dot(x.transpose(), y)


# dropout
def dropout(x, drop_prob=0, seed=123):

    assert 0 <= drop_prob <= 1

    n, p = x.shape

    if drop_prob == 0:
        return x

    if drop_prob == 1:
        return np.zeros_like(x)

    np.random.seed(seed)
    dropped_indices = np.random.rand(n, p) > drop_prob

    return dropped_indices * x / (1 - drop_prob)


# one-hot encoding
@memoize
def one_hot_encode(x_clusters, n_clusters):

    assert (
        max(x_clusters) <= n_clusters
    ), "you must have max(x_clusters) <= n_clusters"

    n_obs = len(x_clusters)
    res = np.zeros((n_obs, n_clusters))

    for i in range(n_obs):
        res[i, x_clusters[i]] = 1

    return res


# one-hot encoding
@memoize
def one_hot_encode2(y, n_classes):
    
    n_obs = len(y)
    
    res = np.zeros((n_obs, n_classes))

    for i in range(n_obs):
        res[i, y[i]] = 1

    return res


# row bind
def rbind(x, y):

    # if len(x.shape) == 1 or len(y.shape) == 1:
    return np.row_stack((x, y))


# from sklearn.utils.exmath
def safe_sparse_dot(a, b, dense_output=False):
    """Dot product that handle the sparse matrix case correctly

    Uses BLAS GEMM as replacement for numpy.dot where possible
    to avoid unnecessary copies.

    Parameters
    ----------
    a : array or sparse matrix
    b : array or sparse matrix
    dense_output : boolean, default False
        When False, either ``a`` or ``b`` being sparse will yield sparse
        output. When True, output will always be an array.

    Returns
    -------
    dot_product : array or sparse matrix
        sparse if ``a`` or ``b`` is sparse and ``dense_output=False``.
    """
    if sparse.issparse(a) or sparse.issparse(b):
        ret = a * b
        if dense_output and hasattr(ret, "toarray"):
            ret = ret.toarray()
        return ret
    else:
        return np.dot(a, b)


# scale... covariates
def scale_covariates(
    X, choice="std", training=True, scaler=None
):

    scaling_options = {
        "std": StandardScaler(
            copy=True, with_mean=True, with_std=True
        ),
        "minmax": MinMaxScaler(),
    }

    if training == True:
        # scaler must be not None
        scaler = scaling_options[choice]
        scaled_X = scaler.fit_transform(X)
        return scaler, scaled_X

    # training == False:
    # scaler must be not None
    return scaler.transform(X)


# from sklearn.utils.exmath
def squared_norm(x):
    """Squared Euclidean or Frobenius norm of x.

    Faster than norm(x) ** 2.

    Parameters
    ----------
    x : array_like

    Returns
    -------
    float
        The Euclidean norm when x is a vector, the Frobenius norm when x
        is a matrix (2-d array).
    """
    x = np.ravel(x, order="K")
    return np.dot(x, x)


# computes x%*%t(y)
def tcrossprod(x, y=None):
    # assert on dimensions
    if y is None:
        return np.dot(x, x.transpose())
    else:
        return np.dot(x, y.transpose())


# convert vector to numpy array
def to_np_array(X):
    return np.array(X.copy(), ndmin=2)


# scale matrix
# def scale_matrix(X, x_means=None, x_vars=None):
#
#    if ((x_means is None) & (x_vars is None)):
#        x_means = X.mean(axis = 0)
#        x_vars = X.var(axis = 0)
#        return ((X - x_means)/np.sqrt(x_vars),
#                x_means,
#                x_vars)
#
#    if ((x_means is not None) & (x_vars is None)):
#        return X - x_means
#
#    if ((x_means is None) & (x_vars is not None)):
#        return X/np.sqrt(x_vars)
#
#    if ((x_means is not None) & (x_vars is not None)):
#        return (X - x_means)/np.sqrt(x_vars)



"""Ridge model for regression"""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
from scipy.optimize import minimize
import sklearn.metrics as skm2
from .ridge import Ridge
from ..utils import matrixops as mo
from ..utils import misc as mx
from sklearn.base import ClassifierMixin
from scipy.special import logsumexp
from scipy.linalg import pinv


class RidgeClassifierMtask(Ridge, ClassifierMixin):
    """Ridge regression model class with 2 regularization parameters derived from class Ridge
    
       Parameters
       ----------
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not 
           (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
               no clustering)
       cluster_encode: bool
           defines how the variable containing clusters is treated (default is one-hot)
           if `False`, then labels are used, without one-hot encoding
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       lambda1: float
           regularization parameter on direct link
       lambda2: float
           regularization parameter on hidden layer
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
    """

    # construct the object -----

    def __init__(
        self,
        n_hidden_features=5,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        lambda1=0.1,
        lambda2=0.1,
        seed=123,
    ):

        super().__init__(
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=1,
            row_sample=1,
            lambda1=lambda1,
            lambda2=lambda2,
            seed=seed,
        )

        self.type_fit = "regression"

    def fit(self, X, y, **kwargs):
        """Fit Ridge model to training data (X, y).           

        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        y: array-like, shape = [n_samples]
               Target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set or self.obj.fit
               
        Returns
        -------
        self: object
        """
        
        assert mx.is_factor(y), "y must contain only integers"                 
        
        output_y, scaled_Z = self.cook_training_set(
            y=y, X=X, **kwargs
        )
        
        n_X, p_X = X.shape
        n_Z, p_Z = scaled_Z.shape
        
        self.n_classes = len(np.unique(y))
        
        # multitask response
        Y = mo.one_hot_encode2(y, self.n_classes)                
        
        if self.n_clusters > 0:
            if self.encode_clusters == True:
                n_features = p_X + self.n_clusters
            else:
                n_features = p_X + 1
        else:
            n_features = p_X

        X_ = scaled_Z[:, 0:n_features]
        Phi_X_ = scaled_Z[:, n_features:p_Z]                

        B = mo.crossprod(X_) + self.lambda1 * np.diag(
            np.repeat(1, X_.shape[1])
        )
        C = mo.crossprod(Phi_X_, X_)
        D = mo.crossprod(Phi_X_) + self.lambda2 * np.diag(
            np.repeat(1, Phi_X_.shape[1])
        )
        B_inv = pinv(B)
        W = np.dot(C, B_inv)
        S_mat = D - mo.tcrossprod(W, C)
        S_inv = pinv(S_mat)
        Y2 = np.dot(S_inv, W)
        inv = mo.rbind(
            mo.cbind(
                B_inv + mo.crossprod(W, Y2), -np.transpose(Y2)
            ),
            mo.cbind(-Y2, S_inv),
        )
            
        self.beta = np.dot(
            inv, mo.crossprod(scaled_Z, Y)
        )

        return self

    def predict(self, X, **kwargs):
        """Predict test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions: {array-like}
        """

        return np.argmax(
            self.predict_proba(X, **kwargs), axis=1
        )

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        probability estimates for test data: {array-like}        
        """        
        
        if len(X.shape) == 1:

            n_features = X.shape[0]
            new_X = mo.rbind(
                X.reshape(1, n_features),
                np.ones(n_features).reshape(1, n_features),
            )

            Z = self.cook_test_set(new_X, **kwargs)

        else:

            Z = self.cook_test_set(X, **kwargs)

        ZB = mo.safe_sparse_dot(Z, self.beta)
        
        exp_ZB = np.exp(ZB)

        return exp_ZB / exp_ZB.sum(axis=1)[:, None]

    def score(self, X, y, scoring=None, **kwargs):
        """ Score the model on test set covariates X and response y. """

        preds = self.predict(X)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)



from setuptools import setup, find_packages
from os import path


with open("README.md", "r") as fh:
    long_description = fh.read()
    
# get the dependencies and installs
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]

setup(
    name="nnetsauce",
    version="0.3.0",
    url="https://github.com/thierrymoudiki/nnetsauce",
    packages=find_packages(),
    author="Thierry Moudiki",
    author_email="thierry.moudiki@gmail.com",
    description="Machine Learning using combinations of Neural Networks' layers",
    download_url="https://github.com/thierrymoudiki/nnetsauce",
    install_requires=["joblib>=0.13.2", "numpy >= 1.13.0", "scipy >= 0.19.0", 
                      "scikit-learn >= 0.18.0", "tqdm>=4.28.1"].append(install_requires),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)



from .nodesimulation import (
    generate_sobol2,
    generate_uniform,
    generate_hammersley,
    generate_halton,
)

from .rowsubsampling import subsample


try:

    from .nodesimulation import generate_sobol_cpp
    from .nodesimulation import generate_halton_cpp

    __all__ = [
        "generate_sobol_cpp",
        "generate_sobol2",
        "generate_uniform",
        "generate_hammersley",
        "generate_halton",
        "generate_halton_cpp",
        "subsample",
    ]

except:

    __all__ = [
        "generate_sobol2",
        "generate_uniform",
        "generate_hammersley",
        "generate_halton",
        "subsample",
    ]



"""Random 'Forest' model"""

# Authors: Thierry Moudiki
#
# License: BSD 3

import numpy as np
import sklearn.metrics as skm2
from .bag import RandomBag
from ..custom import CustomClassifier
from ..utils import misc as mx
from ..utils import Progbar
from sklearn.base import ClassifierMixin
import pickle
from joblib import Parallel, delayed
from tqdm import tqdm


class RandomBagClassifier(RandomBag, ClassifierMixin):
    """Random 'Forest' Classification model class derived from class Boosting
    
       Parameters
       ----------
       obj: object
           any object containing a method fit (obj.fit()) and a method predict 
           (obj.predict())
       n_estimators: int
           number of boosting iterations
       n_hidden_features: int
           number of nodes in the hidden layer
       activation_name: str
           activation function: 'relu', 'tanh', 'sigmoid', 'prelu' or 'elu'
       a: float
           hyperparameter for 'prelu' or 'elu' activation function
       nodes_sim: str
           type of simulation for the nodes: 'sobol', 'hammersley', 'halton', 
           'uniform'
       bias: boolean
           indicates if the hidden layer contains a bias term (True) or not 
           (False)
       dropout: float
           regularization parameter; (random) percentage of nodes dropped out 
           of the training
       direct_link: boolean
           indicates if the original predictors are included (True) in model's 
           fitting or not (False)
       n_clusters: int
           number of clusters for 'kmeans' or 'gmm' clustering (could be 0: 
               no clustering)
       cluster_encode: bool
           defines how the variable containing clusters is treated (default is one-hot)
           if `False`, then labels are used, without one-hot encoding
       type_clust: str
           type of clustering method: currently k-means ('kmeans') or Gaussian 
           Mixture Model ('gmm')
       type_scaling: a tuple of 3 strings
           scaling methods for inputs, hidden layer, and clustering respectively
           (and when relevant). 
           Currently available: standardization ('std') or MinMax scaling ('minmax')
       col_sample: float
           percentage of covariates randomly chosen for training   
       row_sample: float
           percentage of rows chosen for training, by stratified bootstrapping    
       seed: int 
           reproducibility seed for nodes_sim=='uniform'
    """

    # construct the object -----

    def __init__(
        self,
        obj,
        n_estimators=10,
        n_hidden_features=1,
        activation_name="relu",
        a=0.01,
        nodes_sim="sobol",
        bias=True,
        dropout=0,
        direct_link=False,
        n_clusters=2,
        cluster_encode=True,
        type_clust="kmeans",
        type_scaling=("std", "std", "std"),
        col_sample=1,
        row_sample=1,
        n_jobs=None,
        seed=123,
        verbose=1,
    ):

        super().__init__(
            obj=obj,
            n_estimators=n_estimators,
            n_hidden_features=n_hidden_features,
            activation_name=activation_name,
            a=a,
            nodes_sim=nodes_sim,
            bias=bias,
            dropout=dropout,
            direct_link=direct_link,
            n_clusters=n_clusters,
            cluster_encode=cluster_encode,
            type_clust=type_clust,
            type_scaling=type_scaling,
            col_sample=col_sample,
            row_sample=row_sample,
            seed=seed,
        )

        self.type_fit = "classification"
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.voter = dict.fromkeys(range(n_estimators))

    def fit(self, X, y, **kwargs):
        """Fit Random 'Forest' model to training data (X, y).
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        y: array-like, shape = [n_samples]
               Target values.
    
        **kwargs: additional parameters to be passed to 
                  self.cook_training_set or self.obj.fit
               
        Returns
        -------
        self: object
        """

        assert mx.is_factor(
            y.tolist()
        ), "y must contain only integers"

        # training
        n, p = X.shape
        self.n_classes = len(np.unique(y))

        base_learner = CustomClassifier(
            self.obj,
            n_hidden_features=self.n_hidden_features,
            activation_name=self.activation_name,
            a=self.a,
            nodes_sim=self.nodes_sim,
            bias=self.bias,
            dropout=self.dropout,
            direct_link=self.direct_link,
            n_clusters=self.n_clusters,
            type_clust=self.type_clust,
            type_scaling=self.type_scaling,
            col_sample=self.col_sample,
            row_sample=self.row_sample,
            seed=self.seed,
        )

        if self.verbose == 1:
            pbar = Progbar(self.n_estimators)

        # 1 - Sequential training -----

        if self.n_jobs is None:

            for idx, m in enumerate(
                range(self.n_estimators)
            ):

                try:

                    base_learner.fit(X, y, **kwargs)

                    self.voter.update(
                        {
                            m: pickle.loads(
                                pickle.dumps(
                                    base_learner, -1
                                )
                            )
                        }
                    )

                    base_learner.set_params(
                        seed=self.seed + (m + 1) * 1000
                    )

                    if self.verbose == 1:
                        pbar.update(m)

                except:

                    if self.verbose == 1:
                        pbar.update(m)

                    continue

            if self.verbose == 1:
                pbar.update(self.n_estimators)

            self.n_estimators = len(self.voter)

            return self

        # 2 - Parallel training -----

        # if self.n_jobs is not None:
        def fit_estimators(m):
            base_learner = CustomClassifier(
                self.obj,
                n_hidden_features=self.n_hidden_features,
                activation_name=self.activation_name,
                a=self.a,
                nodes_sim=self.nodes_sim,
                bias=self.bias,
                dropout=self.dropout,
                direct_link=self.direct_link,
                n_clusters=self.n_clusters,
                type_clust=self.type_clust,
                type_scaling=self.type_scaling,
                col_sample=self.col_sample,
                row_sample=self.row_sample,
                seed=self.seed + m * 1000,
            )
            base_learner.fit(X, y, **kwargs)
            self.voter.update(
                {
                    m: pickle.loads(
                        pickle.dumps(base_learner, -1)
                    )
                }
            )

        if self.verbose == 1:

            Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(fit_estimators)(m)
                for m in tqdm(range(self.n_estimators))
            )
        else:

            Parallel(n_jobs=self.n_jobs, prefer="threads")(
                delayed(fit_estimators)(m)
                for m in range(self.n_estimators)
            )

        self.n_estimators = len(self.voter)

        return self

    def predict(self, X, weights=None, **kwargs):
        """Predict test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        model predictions: {array-like}        
        """
        return self.predict_proba(
            X, weights, **kwargs
        ).argmax(axis=1)

    def predict_proba(self, X, weights=None, **kwargs):
        """Predict probabilities for test data X.
        
        Parameters
        ----------
        X: {array-like}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number 
            of samples and n_features is the number of features.
        
        **kwargs: additional parameters to be passed to 
                  self.cook_test_set
               
        Returns
        -------
        probability estimates for test data: {array-like}        
        """

        def calculate_probas(
            voter, weights=None, verbose=None
        ):

            ensemble_proba = 0

            n_iter = len(voter)

            assert (
                n_iter > 0
            ), "no estimator found in `RandomBag` ensemble"

            if weights is None:

                for idx, elt in voter.items():

                    ensemble_proba += elt.predict_proba(X)

                    if verbose == 1:
                        pbar.update(idx)

                if verbose == 1:
                    pbar.update(n_iter)

                return ensemble_proba / n_iter

            # if weights is not None:
            for idx, elt in voter.items():

                ensemble_proba += weights[
                    idx
                ] * elt.predict_proba(X)

                if verbose == 1:
                    pbar.update(idx)

            if verbose == 1:
                pbar.update(n_iter)

            return ensemble_proba

        # end calculate_probas ----

        if self.n_jobs is None:

            if self.verbose == 1:
                pbar = Progbar(self.n_estimators)

            if weights is None:

                return calculate_probas(
                    self.voter, verbose=self.verbose
                )

            # if weights is not None:
            self.weights = weights

            return calculate_probas(
                self.voter, weights, verbose=self.verbose
            )

        # if self.n_jobs is not None:
        def predict_estimator(m):
            return self.voter[m].predict_proba(X)

        if self.verbose == 1:

            preds = Parallel(
                n_jobs=self.n_jobs, prefer="threads"
            )(
                delayed(predict_estimator)(m)
                for m in tqdm(range(self.n_estimators))
            )

        else:

            preds = Parallel(
                n_jobs=self.n_jobs, prefer="threads"
            )(
                delayed(predict_estimator)(m)
                for m in range(self.n_estimators)
            )

        ensemble_proba = 0

        if weights is None:

            for i in range(self.n_estimators):

                ensemble_proba += preds[i]

            return ensemble_proba / self.n_estimators

        for i in range(self.n_estimators):

            ensemble_proba += weights[i] * preds[i]

        return ensemble_proba

    def score(
        self, X, y, weights=None, scoring=None, **kwargs
    ):
        """ Score the model on test set covariates X and response y. """

        preds = self.predict(X, weights, **kwargs)

        if scoring is None:
            scoring = "accuracy"

        # check inputs
        assert scoring in (
            "accuracy",
            "average_precision",
            "brier_score_loss",
            "f1",
            "f1_micro",
            "f1_macro",
            "f1_weighted",
            "f1_samples",
            "neg_log_loss",
            "precision",
            "recall",
            "roc_auc",
        ), "'scoring' should be in ('accuracy', 'average_precision', \
                           'brier_score_loss', 'f1', 'f1_micro', \
                           'f1_macro', 'f1_weighted',  'f1_samples', \
                           'neg_log_loss', 'precision', 'recall', \
                           'roc_auc')"

        scoring_options = {
            "accuracy": skm2.accuracy_score,
            "average_precision": skm2.average_precision_score,
            "brier_score_loss": skm2.brier_score_loss,
            "f1": skm2.f1_score,
            "f1_micro": skm2.f1_score,
            "f1_macro": skm2.f1_score,
            "f1_weighted": skm2.f1_score,
            "f1_samples": skm2.f1_score,
            "neg_log_loss": skm2.log_loss,
            "precision": skm2.precision_score,
            "recall": skm2.recall_score,
            "roc_auc": skm2.roc_auc_score,
        }

        return scoring_options[scoring](y, preds, **kwargs)
